"""Module containing the class definition for a steel cross-section element."""

from dataclasses import dataclass

from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import KG_M, MM, MPA
from blueprints.unit_conversion import MM2_TO_M2


@dataclass(frozen=True, kw_only=True)
class SteelElement:
    """
    General class for a steel cross-section element.

    Parameters
    ----------
    cross_section : CrossSection
        The cross-section of the steel element.
    material : SteelMaterial
        The material of the steel element.
    nominal_thickness : MM
        The nominal thickness of the steel element.

        This is used to calculate the yield and ultimate strength of the steel element.
        But be aware that there is no internal check to make sure that the given nominal thickness of this steel element
        is actually the same thickness of the cross-section.
    """

    cross_section: CrossSection
    material: SteelMaterial
    nominal_thickness: MM

    @property
    def weight_per_meter(self) -> KG_M:
        """
        Calculate the weight per meter of the steel element.

        Returns
        -------
        KG_M
            The weight per meter of the steel element.
        """
        return self.material.density * (self.cross_section.area * MM2_TO_M2)

    @property
    def yield_strength(self) -> MPA:
        """
        Calculate the yield strength of the steel element.

        Returns
        -------
        MPa
            The yield strength of the steel element.
        """
        fy = self.material.yield_strength(thickness=self.nominal_thickness)
        if fy is None:
            raise ValueError("Yield strength is not defined for this material.")
        return fy

    @property
    def ultimate_strength(self) -> MPA:
        """
        Calculate the ultimate strength of the steel element.

        Returns
        -------
        MPa
            The ultimate strength of the steel element.
        """
        fu = self.material.ultimate_strength(thickness=self.nominal_thickness)
        if fu is None:
            raise ValueError("Ultimate strength is not defined for this material.")
        return fu



"""Testing formula 8.16 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_16 import (
    Form8Dot16BasicTransmissionLength,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm8Dot16BasicTransmissionLength:
    """Validation for formula 8.16 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 0.25  # [-]
        diameter = 8  # mm
        sigma_pm0 = 350  # MPa
        f_bpt = 5  # MPa
        form_8_16 = Form8Dot16BasicTransmissionLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            diameter=diameter,
            sigma_pm0=sigma_pm0,
            f_bpt=f_bpt,
        )

        # manually calculated result
        manually_calculated_result = 140  # mm

        assert form_8_16 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_negative_alpha_1(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for alpha_1."""
        alpha_1 = -1
        alpha_2 = 1
        diameter = 8
        sigma_pm0 = 0.5
        f_bpt = 5
        with pytest.raises(NegativeValueError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    def test_raise_error_if_negative_alpha_2(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for alpha_2."""
        alpha_1 = 1
        alpha_2 = -1
        diameter = 8
        sigma_pm0 = 0.5
        f_bpt = 5
        with pytest.raises(NegativeValueError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    def test_raise_error_if_negative_diameter(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for diameter."""
        alpha_1 = 1
        alpha_2 = 1
        diameter = -8
        sigma_pm0 = 0.5
        f_bpt = 5
        with pytest.raises(NegativeValueError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    def test_raise_error_if_negative_sigma_pm0(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for sigma_pm0."""
        alpha_1 = 1
        alpha_2 = 1
        diameter = 8
        sigma_pm0 = -0.5
        f_bpt = 5
        with pytest.raises(NegativeValueError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    def test_raise_error_if_negative_f_bpt(self) -> None:
        """Test that a LessOrEqualToZeroError is raised when a negative value is passed for f_bpt."""
        alpha_1 = 1
        alpha_2 = 1
        diameter = 8
        sigma_pm0 = 0.5
        f_bpt = -5
        with pytest.raises(LessOrEqualToZeroError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    def test_raise_error_if_zero_f_bpt(self) -> None:
        """Test that a NegativeValueError is raised when a zero value is passed for f_bpt."""
        alpha_1 = 1
        alpha_2 = 1
        diameter = 8
        sigma_pm0 = 0.5
        f_bpt = 0
        with pytest.raises(LessOrEqualToZeroError):
            Form8Dot16BasicTransmissionLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                diameter=diameter,
                sigma_pm0=sigma_pm0,
                f_bpt=f_bpt,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (
                    r"l_{pt} = \alpha_1 \cdot \alpha_2 \cdot Ø \cdot \frac{\sigma_{pm0}}{f_{bpt}} = "
                    r"1.00 \cdot 0.25 \cdot 8.00 \cdot \frac{350.00}{5.00} = 140.00"
                ),
            ),
            ("short", r"l_{pt} = 140.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 0.25  # [-]
        diameter = 8  # mm
        sigma_pm0 = 350  # MPa
        f_bpt = 5  # MPa

        # Object to test
        form_8_16_latex = Form8Dot16BasicTransmissionLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            diameter=diameter,
            sigma_pm0=sigma_pm0,
            f_bpt=f_bpt,
        ).latex()

        actual = {"complete": form_8_16_latex.complete, "short": form_8_16_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 7.3 from EN 1995-1-1:2004."""

from blueprints.codes.eurocode.en_1995_1_1_2004 import EN_1995_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, KN, MM, MM_KN
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form7Dot3RatioDeflectionPointLoadUC(Formula):
    r"""
    Class representing formula 7.3 for the calculation of ratio between maximum instantaneous
    vertical deflection due to vertical static point load F and the applied force F.
    """

    label = "7.3"
    source_document = EN_1995_1_1_2004

    def __init__(self, w: MM, f: KN, alpha: MM_KN) -> None:
        r"""
        [\frac{w/F}{\alpha}] Unity check of ratio between maximum instantaneous vertical deflection caused by a
        vertical static point load F and the applied force F.

        EN 1995-1-1:2004 art 7.3.3(2) - Formula (7.3)

        Parameters
        ----------
        w : MM
            [$w$] is the maximum instantaneous vertical deflection caused by a vertical static point load F [$mm$].
        f : KN
            [$F$] Vertical static point load F that can engage in any point of the floor [$kN$].
        alpha : MM_KN
            [$\alpha$] Limit of ratio between instantaneous deflection and point load (equal to 1 according to Dutch National Annex) [$mm/kN$].

        Returns
        -------
        None
        """
        super().__init__()
        self.w = w
        self.f = f
        self.alpha = alpha

    @staticmethod
    def _evaluate(w: MM, f: KN, alpha: MM_KN) -> DIMENSIONLESS:
        """Evaluate the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(w=w, f=f, alpha=alpha)
        ratio = w / f
        return ratio / alpha

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.3."""
        eq_for: str = r"\frac{w/F}{\alpha}"
        repl_symb = {
            "w": f"{self.w:.{n}f}",
            "F": f"{self.f:.{n}f}",
            r"\alpha": f"{self.alpha:.{n}f}",
        }
        return LatexFormula(
            return_symbol=r"UC",
            result=f"{self:.{n}f}",
            equation=eq_for,
            numeric_equation=latex_replace_symbols(eq_for, repl_symb),
        )



"""Testing formula 6.38 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_38 import Form6Dot38MaxShearStress
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot38MaxShearStress:
    """Validation for formula 6.38 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        beta = 1.2
        v_ed = 500.0
        u_i = 300.0
        d = 200.0

        # Object to test
        formula = Form6Dot38MaxShearStress(beta=beta, v_ed=v_ed, u_i=u_i, d=d)

        # Expected result, manually calculated
        manually_calculated_result = 0.01  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("beta", "v_ed", "u_i", "d"),
        [
            (-1.2, 500.0, 300.0, 200.0),  # beta is negative
            (1.2, -500.0, 300.0, 200.0),  # v_ed is negative
            (1.2, 500.0, -300.0, 200.0),  # u_i is negative
            (1.2, 500.0, 300.0, -200.0),  # d is negative
            (1.2, 500.0, 0.0, 200.0),  # u_i is zero
            (1.2, 500.0, 300.0, 0.0),  # d is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, beta: float, v_ed: float, u_i: float, d: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot38MaxShearStress(beta=beta, v_ed=v_ed, u_i=u_i, d=d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"v_{Ed} = \beta \cdot \frac{V_{Ed}}{u_{i} \cdot d} = 1.200 \cdot \frac{500.000}{300.000 \cdot 200.000} = 0.010 \ MPa",
            ),
            ("short", r"v_{Ed} = 0.010 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta = 1.2
        v_ed = 500.0
        u_i = 300.0
        d = 200.0

        # Object to test
        latex = Form6Dot38MaxShearStress(beta=beta, v_ed=v_ed, u_i=u_i, d=d).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.78 and 6.79 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, N
from blueprints.validations import raise_if_negative


class Form6Dot78And79FatigueResistance(Formula):
    r"""Class representing formula 6.78 and 6.79 for the calculation of the fatigue resistance."""

    label = "6.78/6.79"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        v_ed_max: N,
        v_ed_min: N,
        v_rd_c: N,
        f_ck: DIMENSIONLESS,
    ) -> None:
        r"""[$\sigma_{Rd,max}$] Fatigue resistance check [$-$].

        EN 1992-1-1:2004 art.6.8.7(4) - Formula (6.78 and 6.79)

        Parameters
        ----------
        v_ed_max : N
            [$V_{Ed,max}$] Design value of the maximum applied shear force under frequent load combination [$N$].
        v_ed_min : N
            [$V_{Ed,min}$] Design value of the minimum applied shear force under frequent load combination
            in the cross-section where [$V_{Ed,max}$] occurs [$N$].
        v_rd_c : N
            [$V_{Rd,c}$] Design value for shear-resistance according to Expression (6.2.a) [$N$].
        f_ck : DIMENSIONLESS
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.v_ed_max = v_ed_max
        self.v_ed_min = v_ed_min
        self.v_rd_c = v_rd_c
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        v_ed_max: N,
        v_ed_min: N,
        v_rd_c: N,
        f_ck: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck, v_rd_c=v_rd_c)

        abs_v_ed_max = abs(v_ed_max)
        abs_v_ed_min = abs(v_ed_min)
        abs_v_rd_c = abs(v_rd_c)

        # the spirit of the equation in the Eurocode is to differentiate between the two cases, not to throw an error
        v_ed_max_non_zero = 1e-12 if v_ed_max == 0 else v_ed_max

        if v_ed_min / v_ed_max_non_zero >= 0:
            output = abs_v_ed_max / abs_v_rd_c <= min(0.5 + 0.45 * abs_v_ed_min / abs_v_rd_c, 0.9 if f_ck <= 50 else 0.8)
        else:
            output = abs_v_ed_max / abs_v_rd_c <= 0.5 - abs_v_ed_min / abs_v_rd_c

        return output

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.78/6.79."""
        _equation: str = (
            r"\begin{cases} \frac{\left|V_{Ed,max}\right|}{\left|V_{Rd,c}\right|} \leq "
            r"\min\left(0.5 + 0.45 \cdot \frac{\left|V_{Ed,min}\right|}{\left|V_{Rd,c}\right|}, "
            r"\begin{cases} 0.9 & \text{if } f_{ck} \le 50 \\ 0.8 & \text{if } f_{ck} > 50 \end{cases} \right) & \text{if } "
            r"\frac{V_{Ed,min}}{V_{Ed,max}} \geq 0 \\ \frac{\left|V_{Ed,max}\right|}{\left|V_{Rd,c}\right|} \leq 0.5 - "
            r"\frac{\left|V_{Ed,min}\right|}{\left|V_{Rd,c}\right|} & "
            r"\text{if } \frac{V_{Ed,min}}{V_{Ed,max}} < 0 \end{cases}"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "V_{Ed,max}": f"{self.v_ed_max:.{n}f}",
                "V_{Ed,min}": f"{self.v_ed_min:.{n}f}",
                "V_{Rd,c}": f"{self.v_rd_c:.{n}f}",
                "f_{ck}": f"{self.f_ck:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 5.28 and 6.5 from EN 1992-1-1:2004: Chapter 5 and 6 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA, N
from blueprints.validations import raise_if_negative


class Form6Dot5ShearForceCheck(Formula):
    r"""Class representing formula 6.5 for the shear force check, [$V_{Ed}$]."""

    label = "6.5"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        v_ed: N,
        b_w: MM,
        d: MM,
        nu: DIMENSIONLESS,
        f_cd: MPA,
    ) -> None:
        r"""[$V_{Ed}$] Shear force check [$N$].

        EN 1992-1-1:2004 art.6.2.2(6) - Formula (6.5)

        Parameters
        ----------
        v_ed : N
            [$V_{Ed}$] Design value of shear force [$N$].
        b_w : MM
            [$b_w$] Width of the web [$mm$].
        d : MM
            [$d$] Effective depth [$mm$].
        nu : DIMENSIONLESS
            [$\nu$] Strength reduction factor for concrete cracked in shear [$-$].
        f_cd : MPA
            [$f_{cd}$] Design value of concrete compressive strength [$MPa$].
        """
        super().__init__()
        self.v_ed = v_ed
        self.b_w = b_w
        self.d = d
        self.nu = nu
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        v_ed: N,
        b_w: MM,
        d: MM,
        nu: DIMENSIONLESS,
        f_cd: MPA,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            v_ed=v_ed,
            b_w=b_w,
            d=d,
            nu=nu,
            f_cd=f_cd,
        )
        return v_ed <= 0.5 * b_w * d * nu * f_cd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.38a."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"V_{Ed} \leq 0.5 \cdot b_w \cdot d \cdot \nu \cdot f_{cd}",
            numeric_equation=rf"{self.v_ed:.{n}f} \leq 0.5 \cdot {self.b_w:.{n}f} \cdot {self.d:.{n}f} \cdot {self.nu:.{n}f} \cdot {self.f_cd:.{n}f}",
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 7.5 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_5 import Form7Dot5AdjustedBondStrengthRatio
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot5AdjustedBondStrengthRatio:
    """Validation for formula 7.5 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        xi = 1.2
        diam_s = 20.0
        diam_p = 15.0

        # Object to test
        formula = Form7Dot5AdjustedBondStrengthRatio(xi=xi, diam_s=diam_s, diam_p=diam_p)

        # Expected result, manually calculated
        manually_calculated_result = 1.26491106407  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("xi", "diam_s", "diam_p"),
        [
            (-1.2, 20.0, 15.0),  # xi is negative
            (1.2, -20.0, 15.0),  # diam_s is negative
            (1.2, 20.0, -15.0),  # diam_p is negative
            (1.2, 20.0, 0.0),  # diam_p is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, xi: float, diam_s: float, diam_p: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot5AdjustedBondStrengthRatio(xi=xi, diam_s=diam_s, diam_p=diam_p)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\xi_1 = \sqrt{\xi \cdot \left( \frac{⌀_s}{⌀_p} \right)} = \sqrt{1.200 \cdot \left( \frac{20.000}{15.000} \right)} = 1.265 \ -",
            ),
            ("short", r"\xi_1 = 1.265 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        xi = 1.2
        diam_s = 20.0
        diam_p = 15.0

        # Object to test
        latex = Form7Dot5AdjustedBondStrengthRatio(xi=xi, diam_s=diam_s, diam_p=diam_p).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing sub-formula 2 of 3.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_14 import SubForm3Dot14K


class TestSub2Form3Dot14K:
    """Validation for sub-formula 2 of 3.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        e_cm = 3.3  # MPa
        epsilon_c1 = 0.43  # -
        f_cm = 6.7  # MPa

        sub_2_form_3_14 = SubForm3Dot14K(e_cm=e_cm, epsilon_c1=epsilon_c1, f_cm=f_cm)

        # Expected result, manually calculated
        manually_calculated_result = 0.22238

        assert sub_2_form_3_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_e_cm_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        e_cm = -3.3  # MPa
        epsilon_c1 = 0.43  # -
        f_cm = 6.7  # MPa

        with pytest.raises(ValueError):
            SubForm3Dot14K(e_cm=e_cm, epsilon_c1=epsilon_c1, f_cm=f_cm)

    def test_raise_error_when_negative_or_zero_f_cm_is_given(self) -> None:
        """Test formula raising error by a negative or zero value."""
        # Example values
        e_cm = 3.3  # MPa
        epsilon_c1 = 0.43  # -
        f_cm = 0  # MPa

        with pytest.raises(ValueError):
            SubForm3Dot14K(e_cm=e_cm, epsilon_c1=epsilon_c1, f_cm=f_cm)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"k = 1.05 \cdot E_{cm} \cdot |\epsilon_{c1}| / f_{cm} = 1.05 \cdot 3.300 \cdot |0.430| / 6.700 = 0.222",
            ),
            ("short", r"k = 0.222"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_cm = 3.3  # MPa
        epsilon_c1 = 0.43  # -
        f_cm = 6.7  # MPa

        # Object to test
        form_3_14sub2_latex = SubForm3Dot14K(e_cm=e_cm, epsilon_c1=epsilon_c1, f_cm=f_cm).latex()

        actual = {"complete": form_3_14sub2_latex.complete, "short": form_3_14sub2_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_4 import Form6Dot4ShearResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot4ShearResistance:
    """Validation for formula 6.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        i = 1000.0
        b_w = 300.0
        s = 500.0
        f_ctd = 2.5
        alpha_l = 1.0
        sigma_cp = 1.5

        # Object to test
        formula = Form6Dot4ShearResistance(i=i, b_w=b_w, s=s, f_ctd=f_ctd, alpha_l=alpha_l, sigma_cp=sigma_cp)

        # Expected result, manually calculated
        manually_calculated_result = 1897.3665961010277

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("i", "b_w", "s", "f_ctd", "alpha_l", "sigma_cp"),
        [
            (-1000.0, 300.0, 500.0, 2.5, 1.0, 1.5),  # i is negative
            (1000.0, -300.0, 500.0, 2.5, 1.0, 1.5),  # b_w is negative
            (1000.0, 300.0, -500.0, 2.5, 1.0, 1.5),  # s is negative
            (1000.0, 300.0, 500.0, -2.5, 1.0, 1.5),  # f_ctd is negative
            (1000.0, 300.0, 500.0, 2.5, -1.0, 1.5),  # alpha_l is negative
            (1000.0, 300.0, 500.0, 2.5, 1.0, -1.5),  # sigma_cp is negative
            (1000.0, 300.0, 0.0, 2.5, 1.0, 1.5),  # s is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, i: float, b_w: float, s: float, f_ctd: float, alpha_l: float, sigma_cp: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot4ShearResistance(i=i, b_w=b_w, s=s, f_ctd=f_ctd, alpha_l=alpha_l, sigma_cp=sigma_cp)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Rd,c} = \frac{I \cdot b_w}{S} \cdot \sqrt{(f_{ctd})^2 + \alpha_l \cdot \sigma_{cp} \cdot f_{ctd}} = "
                r"\frac{1000.000 \cdot 300.000}{500.000} \cdot \sqrt{(2.500)^2 + 1.000 \cdot 1.500 \cdot 2.500} = 1897.367 \ N",
            ),
            ("short", r"V_{Rd,c} = 1897.367 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        i = 1000.0
        b_w = 300.0
        s = 500.0
        f_ctd = 2.5
        alpha_l = 1.0
        sigma_cp = 1.5

        # Object to test
        latex = Form6Dot4ShearResistance(i=i, b_w=b_w, s=s, f_ctd=f_ctd, alpha_l=alpha_l, sigma_cp=sigma_cp).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.33 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_33 import Form6Dot33CheckAxialForceY
from blueprints.validations import NegativeValueError


class TestForm6Dot33CheckAxialForceY:
    """Validation for formula 6.33 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        n_ed = 50.0
        n_pl_rd = 300.0

        formula = Form6Dot33CheckAxialForceY(n_ed=n_ed, n_pl_rd=n_pl_rd)

        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        "n_pl_rd",
        [-300.0],  # n_pl_rd is negative
    )
    def test_raise_error_when_invalid_values_are_given(self, n_pl_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot33CheckAxialForceY(n_ed=50.0, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to N_{Ed} \leq 0.25 \cdot N_{pl,Rd} \to 50.000 \leq 0.25 \cdot 300.000 \to OK",
            ),
            (
                "complete_with_units",
                r"CHECK \to N_{Ed} \leq 0.25 \cdot N_{pl,Rd} \to 50.000 \ N \leq 0.25 \cdot 300.000 \ N \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        n_ed = 50.0
        n_pl_rd = 300.0

        latex = Form6Dot33CheckAxialForceY(n_ed=n_ed, n_pl_rd=n_pl_rd).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing sub-formula 2 from 8.15 from EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_4 import Form3Dot4DevelopmentTensileStrength
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_15 import (
    SubForm8Dot15TensileStrengthAtRelease,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestSubForm8Dot15TensileStrengthAtRelease:
    """Validation for sub-formula 2 from 8.15 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_ct = 1.0  # [-]
        f_ctm_t = 2.5  # [MPa]
        gamma_c = 1.5  # [-]
        sub_form_8_15 = SubForm8Dot15TensileStrengthAtRelease(
            alpha_ct=alpha_ct,
            f_ctm_t=f_ctm_t,
            gamma_c=gamma_c,
        )

        # manually calculated result
        manually_calculated_result = 1.166666666666  # [MPa]

        assert sub_form_8_15 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_negative_alpha_ct(self) -> None:
        """Test if NegativeValueError is raised for negative alpha_ct."""
        # example values
        alpha_ct = -1.0  # [-]
        f_ctm_t = 2.5  # [MPa]
        gamma_c = 1.5  # [-]
        with pytest.raises(NegativeValueError):
            SubForm8Dot15TensileStrengthAtRelease(
                alpha_ct=alpha_ct,
                f_ctm_t=f_ctm_t,
                gamma_c=gamma_c,
            )

    def test_negative_f_ctd_t(self) -> None:
        """Test if NegativeValueError is raised for negative f_ctd_t."""
        # example values
        alpha_ct = 1.0  # [-]
        f_ctm_t = -2.5  # [MPa]
        gamma_c = 1.5  # [-]
        with pytest.raises(NegativeValueError):
            SubForm8Dot15TensileStrengthAtRelease(
                alpha_ct=alpha_ct,
                f_ctm_t=f_ctm_t,
                gamma_c=gamma_c,
            )

    def test_negative_gamma_c(self) -> None:
        """Test if LessOrEqualToZeroError is raised for negative gamma_c."""
        # example values
        alpha_ct = 1.0  # [-]
        f_ctm_t = 2.5  # [MPa]
        gamma_c = -1.5  # [-]
        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot15TensileStrengthAtRelease(
                alpha_ct=alpha_ct,
                f_ctm_t=f_ctm_t,
                gamma_c=gamma_c,
            )

    def test_zero_gamma_c(self) -> None:
        """Test if LessOrEqualToZeroError is raised for zero gamma_c."""
        # example values
        alpha_ct = 1.0  # [-]
        f_ctm_t = 2.5  # [MPa]
        gamma_c = 0.0  # [-]
        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot15TensileStrengthAtRelease(
                alpha_ct=alpha_ct,
                f_ctm_t=f_ctm_t,
                gamma_c=gamma_c,
            )

    def test_integration_with_form_3_4(self) -> None:
        """Test the integration with Form3Dot4DevelopmentTensileStrength."""
        # example values
        alpha_ct = 1.0  # [-]
        gamma_c = 1.5  # [-]
        beta_cc_t = 0.7  # [-]
        alpha = 1.0  # [-]
        f_ctm = 2.5  # [MPa]
        f_ctm_t = Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

        sub_form_8_15 = SubForm8Dot15TensileStrengthAtRelease(
            alpha_ct=alpha_ct,
            f_ctm_t=f_ctm_t,
            gamma_c=gamma_c,
        )

        # manually calculated result
        manually_calculated_result = 0.816666666  # [MPa]

        assert sub_form_8_15 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (r"f_{ctd}(t) = \frac{\alpha_{ct} \cdot 0.7 \cdot f_{ctm}(t)}{\gamma_c} = \frac{1.00 \cdot 0.7 \cdot 1.75}{1.50} = 0.82"),
            ),
            ("short", r"f_{ctd}(t) = 0.82"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        alpha_ct = 1.0  # [-]
        gamma_c = 1.5  # [-]
        beta_cc_t = 0.7  # [-]
        alpha = 1.0  # [-]
        f_ctm = 2.5  # [MPa]
        f_ctm_t = Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

        # Object to test
        form_8_15_p2_latex = SubForm8Dot15TensileStrengthAtRelease(
            alpha_ct=alpha_ct,
            f_ctm_t=f_ctm_t,
            gamma_c=gamma_c,
        ).latex()

        actual = {"complete": form_8_15_p2_latex.complete, "short": form_8_15_p2_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test for formula 2.21 from CUR 228."""

import pytest

from blueprints.codes.cur.cur_228.formula_2_21 import Form2Dot21ModulusHorizontalSubgrade
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm2Dot21ModulusHorizontalSubgrade:
    """Validation for formula 2.21 from CUR 228."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        r = 0.5  # m
        e_p = 2.47  # kN/m²
        alpha = 1 / 3  # -
        form_2_21 = Form2Dot21ModulusHorizontalSubgrade(r=r, e_p=e_p, alpha=alpha)

        # Expected result, manually calculated
        manually_calculated_result = 9.187357198

        assert form_2_21 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("r", "alpha"),
        [
            (-0.6, -0.33),
            (0.2, -0.33),
            (0.2, 0.0),
            (0.0, 0.0),
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, r: float, alpha: float) -> None:
        """Test invalid values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form2Dot21ModulusHorizontalSubgrade(r=r, e_p=500, alpha=alpha)

    @pytest.mark.parametrize(
        ("e_p"),
        [
            (-500.0),
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, e_p: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form2Dot21ModulusHorizontalSubgrade(r=0.4, e_p=e_p, alpha=0.33)

    @pytest.mark.parametrize(
        ("r"),
        [
            (0.2),
            (0.29),
        ],
    )
    def test_raise_error_when_invalid_diameter_values_are_given(self, r: float) -> None:
        """Test invalid values."""
        with pytest.raises(ValueError):
            Form2Dot21ModulusHorizontalSubgrade(r=r, e_p=500, alpha=0.33)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (
                    "k_{h} = \\frac{1}{3 \\cdot E_{p}} \\cdot \\left[1.3 \\cdot R_{0} \\left( 2.65 \\frac{R}{R_0}\\right)^\\alpha + \\alpha \\cdot  "
                    "R \\right] = \\frac{1}{3 \\cdot 2.5} \\cdot\\left[1.3 \\cdot 0.3 \\left( 2.65 \\cdot \\frac{0.5}{0.3}\\right)^{0.33}+ 0.33 "
                    "\\cdot 0.5\\right] = 9.19 \\ kN/m^3"
                ),
            ),
            ("short", r"k_{h} = 9.19 \ kN/m^3"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        r = 0.5  # m
        e_p = 2.47  # kN/m²
        alpha = 1 / 3  # -

        # Object to test
        form_2_1_a_latex = Form2Dot21ModulusHorizontalSubgrade(r=r, e_p=e_p, alpha=alpha).latex()

        actual = {"complete": form_2_1_a_latex.complete, "short": form_2_1_a_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 7.15 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_greater_than_90, raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot15MaximumCrackSpacing(Formula):
    r"""Class representing formula 7.15 for the calculation of [$s_{r,max}$]."""

    label = "7.15"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        theta: DIMENSIONLESS,
        sr_max_y: MM,
        sr_max_z: MM,
    ) -> None:
        r"""[$s_{r,max}$] Calculation of the maximum crack spacing, where the angle between the axes of principal stress and the direction of the
        reinforcement, for members reinforced in two orthogonal directions, is significant (>15 degrees) [$mm$].

        EN 1992-1-1:2004 art.7.3.4(4) - Formula (7.15)

        Parameters
        ----------
        theta : DIMENSIONLESS
            [$\theta$] Angle between the reinforcement in the y direction and the direction of the principal tensile stress [$degrees$].
        sr_max_y : MM
            [$s_{r,max,y}$] Crack spacing in the y direction [$mm$].
        sr_max_z : MM
            [$s_{r,max,z}$] Crack spacing in the z direction [$mm$].

        """
        super().__init__()
        self.theta = theta
        self.sr_max_y = sr_max_y
        self.sr_max_z = sr_max_z

    @staticmethod
    def _evaluate(
        theta: DIMENSIONLESS,
        sr_max_y: MM,
        sr_max_z: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(sr_max_y=sr_max_y, sr_max_z=sr_max_z)
        raise_if_negative(theta=theta)
        raise_if_greater_than_90(theta=theta)
        if theta <= 15:
            raise ValueError(f"Invalid theta: {theta}. theta must be significant (>15 degrees)")

        return 1 / ((np.cos(np.deg2rad(theta)) / sr_max_y) + (np.sin(np.deg2rad(theta)) / sr_max_z))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.15."""
        _equation: str = r"\frac{1}{\left(\frac{\cos(\theta)}{s_{r,max,y}}\right) + \left(\frac{\sin(\theta)}{s_{r,max,z}}\right)}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\theta": f"{self.theta:.{n}f}",
                r"s_{r,max,y}": f"{self.sr_max_y:.{n}f}",
                r"s_{r,max,z}": f"{self.sr_max_z:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"s_{r,max}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm",
        )



"""Tests for Circular Reinforced Concrete Sections."""

import pytest
from matplotlib import pyplot as plt
from shapely import LineString, Polygon

from blueprints.structural_sections.concrete.rebar import Rebar
from blueprints.structural_sections.concrete.reinforced_concrete_sections.circular import CircularReinforcedCrossSection
from blueprints.structural_sections.concrete.reinforced_concrete_sections.reinforcement_configurations import ReinforcementByQuantity
from blueprints.structural_sections.concrete.stirrups import StirrupConfiguration


class TestCircularReinforcedCrossSection:
    """Tests for the CircularReinforcedCrossSection class."""

    def test_add_stirrup_along_perimeter(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_stirrup_along_perimeter method."""
        stirrup = circular_reinforced_cross_section.add_stirrup_along_perimeter(
            diameter=12,
            distance=150,
            material=circular_reinforced_cross_section.get_present_steel_materials()[0],
        )
        assert stirrup in circular_reinforced_cross_section.stirrups

    def test_add_longitudinal_reinforcement_by_quantity(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_longitudinal_reinforcement_by_quantity method."""
        assert len(circular_reinforced_cross_section.longitudinal_rebars) == 7

    def test_plot(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the plot method."""
        plot = circular_reinforced_cross_section.plot(show=False, center_line_style={"linewidth": 0.85})
        assert isinstance(plot, plt.Figure)

    def test_reinforcement_weight_longitudinal_bars(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight_longitudinal_bars method."""
        expected_weight = 14.98802804  # kg/m
        assert circular_reinforced_cross_section.reinforcement_weight_longitudinal_bars == pytest.approx(expected=expected_weight, rel=1e-2)

    def test_reinforcement_weight_stirrups(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight_stirrups method."""
        expected_weight = 22.96448757  # kg/m
        assert circular_reinforced_cross_section.reinforcement_weight_stirrups == pytest.approx(expected=expected_weight, rel=1e-3)

    def test_reinforcement_weight(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight method."""
        expected_weight = 37.95251561  # kg/m
        assert circular_reinforced_cross_section.reinforcement_weight == pytest.approx(expected=expected_weight, rel=1e-3)

    def test_reinforcement_area_longitudinal_bars(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the reinforcement_area_longitudinal_bars method."""
        expected_area = 1909.302935  # mm²/m
        assert circular_reinforced_cross_section.reinforcement_area_longitudinal_bars == pytest.approx(expected=expected_area, rel=1e-2)

    def test_concrete_volume(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the concrete_volume method."""
        expected_volume = 0.125663706  # m³/m
        assert circular_reinforced_cross_section.concrete_volume == pytest.approx(expected=expected_volume, rel=1e-2)

    def test_weight_per_volume(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the weight_per_volume method."""
        expected_weight_per_volume = 302.0165231  # kg/m³
        assert circular_reinforced_cross_section.weight_per_volume == pytest.approx(expected=expected_weight_per_volume, rel=1e-2)

    def test_add_longitudinal_rebar_wrong_position(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_longitudinal_rebar method with wrong position."""
        with pytest.raises(ValueError):
            circular_reinforced_cross_section.add_longitudinal_rebar(
                rebar=Rebar(
                    diameter=12,
                    x=2500,
                    y=1000,
                    material=circular_reinforced_cross_section.get_present_steel_materials()[0],
                )
            )

    def test_add_stirrup_configuration_wrong_position(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_stirrup_configuration method with wrong position."""
        with pytest.raises(ValueError):
            circular_reinforced_cross_section.add_stirrup_configuration(
                stirrup=StirrupConfiguration(
                    geometry=Polygon([(0, 0), (0, 100), (5000, 5000), (100, 0)]),
                    diameter=8,
                    distance=150,
                    material=circular_reinforced_cross_section.get_present_steel_materials()[0],
                ),
            )

    def test_add_reinforcement_configuration_by_linestring(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_reinforcement_configuration method with a linestring."""
        linestring = LineString([(-100, 100), (100, 0)])
        circular_reinforced_cross_section.add_reinforcement_configuration(
            line=linestring,
            configuration=ReinforcementByQuantity(
                diameter=12,
                n=7,
                material=circular_reinforced_cross_section.get_present_steel_materials()[0],
            ),
        )
        assert len(circular_reinforced_cross_section.longitudinal_rebars) == 14

    def test_rebar_not_in_cross_section(self, circular_reinforced_cross_section: CircularReinforcedCrossSection) -> None:
        """Test the add_longitudinal_rebar method with a rebar not in the cross-section."""
        rebar = Rebar(
            diameter=12,
            x=2500,
            y=1000,
            material=circular_reinforced_cross_section.get_present_steel_materials()[0],
        )
        circular_reinforced_cross_section._single_longitudinal_rebars.append(rebar)  # noqa: SLF001
        with pytest.raises(ValueError):
            _ = circular_reinforced_cross_section.longitudinal_rebars



"""Tests for CoversRectangular class."""

import pytest

from blueprints.structural_sections.concrete.covers import CoversRectangular
from blueprints.validations import NegativeValueError


class TestCoversRectangular:
    """Tests for the CoversRectangular class."""

    def test_get_covers_info_with_all_equal_covers(self) -> None:
        """Test if the method `get_covers_info` returns the correct string when all covers are equal."""
        covers = CoversRectangular(upper=50.0, right=50.0, lower=50.0, left=50.0)
        assert covers.get_covers_info() == "Cover: 50 mm"

    def test_get_covers_info_with_different_covers(self) -> None:
        """Test if the method `get_covers_info` returns the correct string when all covers are different."""
        covers = CoversRectangular(upper=40.0, right=50.0, lower=60.0, left=70.0)
        expected_output = "Cover:\n  upper: 40 mm\n  right: 50 mm\n  lower: 60 mm\n  left: 70 mm"
        assert covers.get_covers_info() == expected_output

    def test_get_covers_info_with_two_equal_covers(self) -> None:
        """Test if the method `get_covers_info` returns the correct string when two covers are equal."""
        # Upper and left covers are equal
        covers = CoversRectangular(upper=50.0, right=60.0, lower=50.0, left=70.0)
        expected_output = "Cover:\n  upper|lower: 50 mm\n  right: 60 mm\n  left: 70 mm"
        assert covers.get_covers_info() == expected_output

        # Upper and right covers are equal adnd lower and left covers are equal
        covers = CoversRectangular(upper=50.0, right=50.0, lower=60.0, left=60.0)
        expected_output = "Cover:\n  upper|right: 50 mm\n  lower|left: 60 mm"
        assert covers.get_covers_info() == expected_output

    def test_get_covers_info_with_three_equal_covers(self) -> None:
        """Test if the method `get_covers_info` returns the correct string when three covers are equal."""
        covers = CoversRectangular(upper=50.0, right=50.0, lower=50.0, left=60.0)
        expected_output = "Cover:\n  upper|right|lower: 50 mm\n  left: 60 mm"
        assert covers.get_covers_info() == expected_output

    @pytest.mark.parametrize(
        "covers",
        [
            (50.0, 50.0, 50.0, -60.0),
            (50.0, 50.0, -50.0, 60.0),
            (50.0, -50.0, 50.0, 60.0),
            (-50.0, 50.0, 50.0, 60.0),
        ],
    )
    def test_covers_negative_value(self, covers: tuple[float, float, float, float]) -> None:
        """Test if the method `validate` raises an error when a cover is negative."""
        with pytest.raises(NegativeValueError):
            CoversRectangular(upper=covers[0], right=covers[1], lower=covers[2], left=covers[3])



"""Formula 5.29 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot29BetaFactor(Formula):
    r"""Class representing formula 5.29 for the calculation of the beta factor, [$\beta$]."""

    label = "5.29"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        c_0: DIMENSIONLESS,
    ) -> None:
        r"""[$\beta$] Factor which depends on the distribution of first order moment [-].

        EN 1992-1-1:2004 art.5.8.8.2(3) - Formula (5.29)

        Parameters
        ----------
        c_0 : float
            Coefficient which depends on the distribution of first order moment [-].
            For instance, c_0 = 8 for a constant first order moment, c_0 = 9.6 for a parabolic and 12 for a symmetric triangular distribution.
        """
        super().__init__()
        self.c_0 = c_0

    @staticmethod
    def _evaluate(
        c_0: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(c_0=c_0)
        return (np.pi**2) / c_0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.29."""
        return LatexFormula(
            return_symbol=r"\beta",
            result=f"{self:.{n}f}",
            equation=r"\frac{\pi^2}{c_0}",
            numeric_equation=rf"\frac{{\pi^2}}{{{self.c_0:.{n}f}}}",
            comparison_operator_label="=",
            unit="-",
        )



"""Formula 1.0.1 from NEN 9997-1+C2:2017: Chapter 1: General rules."""

import numpy as np

from blueprints.codes.eurocode.nen_9997_1_c2_2017 import NEN_9997_1_C2_2017
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction
from blueprints.type_alias import M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form1Dot0Dot1EquivalentPilePointCenterline(Formula):
    r"""Class representing formula 1.0.1 for the calculation of the equivalent pile point centerline [$D_{eq}$] in [$m$]."""

    label = "1.0.1"
    source_document = NEN_9997_1_C2_2017

    def __init__(self, a: M, b: M) -> None:
        r"""[$D_{eq}$] Equivalent pile point centerline.

        NEN 9997-1+C2:2017 art.1.5.2.106a - Formula (1.0.1)

        Parameters
        ----------
        a : M
            [$a$] minor dimension of the largest cross-section at the pile tip [$m$].
        b : M
            [$b$] major dimension of the largest cross-section at the pile tip [$m$].

            Where: b ≤ 1.5 * a
        """
        super().__init__()
        self.a = a
        self.b = b

    @staticmethod
    def _evaluate(
        a: M,
        b: M,
    ) -> M:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(a=a, b=b)
        if b > 1.5 * a:
            raise ValueError("b must be less than or equal to 1.5 * a")
        return 1.13 * a * np.sqrt(b / a)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 1.0.1."""
        return LatexFormula(
            return_symbol=r"D_{eq}",
            result=f"{self:.{n}f}",
            equation=r"1.13 \cdot a \cdot \sqrt{\frac{min(b, 1.5 \cdot a)}{a}}",
            numeric_equation=rf"1.13 \cdot {self.a} \cdot \sqrt{latex_fraction(numerator=self.b, denominator=self.a)}",
            comparison_operator_label="=",
        )



"""Testing formula 8.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_3 import (
    Form8Dot3RequiredAnchorageLength,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_4 import (
    Form8Dot4DesignAnchorageLength,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_6 import (
    Form8Dot6MinimumTensionAnchorage,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_7 import (
    Form8Dot7MinimumCompressionAnchorage,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot4DesignAnchorageLength:
    """Validation for formula 8.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 450  # mm
        l_b_min = 200  # mm
        form_8_4 = Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        )

        # manually calculated result
        manually_calculated_result = 450  # mm

        assert form_8_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_minimum(self) -> None:
        """Test the evaluation of the result if the minimum is reached."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 200  # mm
        l_b_min = 300  # mm
        form_8_4 = Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        )

        # manually calculated result
        manually_calculated_result = 300  # mm

        assert form_8_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_alpha_product_to_small(self) -> None:
        """Test the evaluation of the result if the product of alpha 2,3,5 is to small."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 0.6  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 450  # mm
        l_b_min = 200  # mm
        form_8_4 = Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        )

        # manually calculated result
        manually_calculated_result = 315  # mm

        assert form_8_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_alpha_product_to_small_with_overwrite(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 0.6  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 450  # mm
        l_b_min = 200  # mm
        min_product_alpha_2_3_5 = 0.85  # [-]
        form_8_4 = Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
            min_product_alpha_2_3_5=min_product_alpha_2_3_5,
        )

        # manually calculated result
        manually_calculated_result = 382.5  # mm

        assert form_8_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("alpha_1", "alpha_2", "alpha_3", "alpha_4", "alpha_5"),
        [
            (-1, 1, 1, 1, 1),  # alpha_1 is negative
            (1, -1, 1, 1, 1),  # alpha_2 is negative
            (1, 1, -1, 1, 1),  # alpha_3 is negative
            (1, 1, 1, -1, 1),  # alpha_4 is negative
            (1, 1, 1, 1, -1),  # alpha_5 is negative
        ],
    )
    def test_negative_alpha(self, alpha_1: float, alpha_2: float, alpha_3: float, alpha_4: float, alpha_5: float) -> None:
        """Test the evaluation of the result if one of the alpha values is negative."""
        with pytest.raises(NegativeValueError):
            Form8Dot4DesignAnchorageLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_4=alpha_4,
                alpha_5=alpha_5,
                l_b_rqd=450,  # mm
                l_b_min=200,  # mm
            )

    def test_integration_with_form_8_3(self) -> None:
        """Test the integration with formula 8.3."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_min = 200  # mm
        diameter = 16  # mm
        sigma_sd = 500  # MPa
        f_bd = 2.5  # MPa
        l_b_rqd = Form8Dot3RequiredAnchorageLength(
            diameter=diameter,
            sigma_sd=sigma_sd,
            f_bd=f_bd,
        )

        # manually calculated result
        manually_calculated_result = 800

        assert Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        ) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_form_8_6(self) -> None:
        """Test the integration with formula 8.6."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 450  # mm
        diameter = 16  # mm
        l_b_min = Form8Dot6MinimumTensionAnchorage(l_b_rqd=l_b_rqd, diameter=diameter)

        # manually calculated result
        manually_calculated_result = 450

        assert Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        ) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_form_8_7(self) -> None:
        """Test the integration with formula 8.7."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 200  # mm
        diameter = 32  # mm
        l_b_min = Form8Dot7MinimumCompressionAnchorage(l_b_rqd=l_b_rqd, diameter=diameter)

        # manually calculated result
        manually_calculated_result = 320

        assert Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        ) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"l_{bd} = \max \left\{\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_4 \cdot \alpha_5 \cdot l_{b,rqd};"
                r" l_{b,min}\right\} = \max \left\{1 \cdot 1 \cdot 1 \cdot 1 \cdot 1 \cdot 200.00; 400\right\} = 400.00",
            ),
            ("short", "l_{bd} = 400.00"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test latex representation of the formula."""
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_4 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 200  # mm
        l_b_min = 400  # mm
        latex = Form8Dot4DesignAnchorageLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Tests for cross-section shapes."""

import pytest
from sectionproperties.analysis import Section
from sectionproperties.post.post import SectionProperties

from blueprints.structural_sections.cross_section_rectangle import RectangularCrossSection


class TestRectangularCrossSection:
    """Tests for the RectangularCrossSection class."""

    def test_area(self, rectangular_cross_section: RectangularCrossSection) -> None:
        """Test the area property of the RectangularCrossSection class."""
        assert rectangular_cross_section.area == pytest.approx(expected=20000.0, rel=1e-6)

    def test_invalid_width(self) -> None:
        """Test that an error is raised for invalid width."""
        with pytest.raises(ValueError, match="Width must be a positive value"):
            RectangularCrossSection(name="InvalidWidth", width=-10.0, height=200.0)

    def test_invalid_height(self) -> None:
        """Test that an error is raised for invalid height."""
        with pytest.raises(ValueError, match="Height must be a positive value"):
            RectangularCrossSection(name="InvalidHeight", width=100.0, height=-200.0)

    def test_polygon(self, rectangular_cross_section: RectangularCrossSection) -> None:
        """Test the geometry property of the RectangularCrossSection class."""
        polygon = rectangular_cross_section.polygon
        assert polygon.bounds == pytest.approx(expected=(50.0, 150.0, 150.0, 350.0), rel=1e-6)

    def test_section(self, rectangular_cross_section: RectangularCrossSection) -> None:
        """Test the geometry property of the RectangularCrossSection class."""
        section = rectangular_cross_section.section()
        assert isinstance(section, Section)

    def test_section_properties_geometric(self, rectangular_cross_section_section_properties: SectionProperties) -> None:
        """Test the section properties of the RectangularCrossSection class."""
        assert rectangular_cross_section_section_properties.area == pytest.approx(expected=20000.0, rel=1e-6)
        assert rectangular_cross_section_section_properties.perimeter == pytest.approx(expected=600.0, rel=1e-6)
        assert rectangular_cross_section_section_properties.ixx_g == pytest.approx(expected=1316666666.6, rel=1e-6)

    def test_section_properties_plastic(self, rectangular_cross_section_section_properties: SectionProperties) -> None:
        """Test the section properties of the RectangularCrossSection class."""
        assert rectangular_cross_section_section_properties.syy == pytest.approx(expected=500000.0, rel=1e-6)
        assert rectangular_cross_section_section_properties.sxx == pytest.approx(expected=1000000.0, rel=1e-6)

    def test_geometry(self, rectangular_cross_section: RectangularCrossSection) -> None:
        """Test the geometry property of the RectangularCrossSection class."""
        geometry = rectangular_cross_section.geometry()
        assert geometry is not None



"""Formula 6.75 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot75MaximumCompressiveStressLevel(Formula):
    r"""Class representing formula 6.75 for the calculation of the maximum compressive stress level [$E_{cd,max,equ}$]."""

    label = "6.75"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        sigma_cd_max_equ: MPA,
        f_cd_fat: MPA,
    ) -> None:
        r"""[$E_{cd,max,equ}$] Calculation of the maximum compressive stress level.

        EN 1992-1-1:2004 art.6.8.7(1) - Formula (6.75)

        Parameters
        ----------
        sigma_cd_max_equ : MPA
            [$\sigma_{cd,max,equ}$] Upper stress of the ultimate amplitude for N cycles [$MPa$].
        f_cd_fat : MPA
            [$f_{cd,fat}$] Design fatigue strength of concrete according to (6.76) [$MPa$].
        """
        super().__init__()
        self.sigma_cd_max_equ = sigma_cd_max_equ
        self.f_cd_fat = f_cd_fat

    @staticmethod
    def _evaluate(
        sigma_cd_max_equ: MPA,
        f_cd_fat: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(sigma_cd_max_equ=sigma_cd_max_equ)
        raise_if_less_or_equal_to_zero(f_cd_fat=f_cd_fat)

        return sigma_cd_max_equ / f_cd_fat

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.75."""
        _equation: str = r"\frac{\sigma_{cd,max,equ}}{f_{cd,fat}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\sigma_{cd,max,equ}": f"{self.sigma_cd_max_equ:.{n}f}",
                r"f_{cd,fat}": f"{self.f_cd_fat:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"E_{cd,max,equ}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 5.7a and 5.7b of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_7ab import Form5Dot7abFlangeEffectiveFlangeWidth
from blueprints.validations import NegativeValueError


class TestForm5Dot7abFlangeEffectiveFlangeWidth:
    """Validation for formula 5.7a and 5.7b from EN 1992-1-1:2004."""

    @pytest.mark.parametrize(
        ("b_i", "l_0", "b_eff_i"),
        [
            (0.4, 3.0, 0.38),
            (0.25, 3.0, 0.25),
            (0.8, 1.5, 0.30),
        ],
    )
    def test_evaluation(self, b_i: float, l_0: float, b_eff_i: float) -> None:
        """Test the evaluation of the result."""
        # Object to test
        form_5_7_ab = Form5Dot7abFlangeEffectiveFlangeWidth(b_i=b_i, l_0=l_0)

        # Expected result, manually calculated
        manually_calculated_result = b_eff_i  # M

        assert form_5_7_ab == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("b_i", "l_0"),
        [
            (0.4, -3.0),
            (-0.25, 3.0),
            (-0.8, -1.5),
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self,
        b_i: float,
        l_0: float,
    ) -> None:
        """Test negative values."""
        with pytest.raises(NegativeValueError):
            Form5Dot7abFlangeEffectiveFlangeWidth(b_i=b_i, l_0=l_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"b_{eff,i} = 0.2b_{i}+0.1l_{0} \le 0.2l_{0}\text{ and }b_{eff,i}\le b_{i} = "
                r"0.2\cdot0.4+0.1\cdot3.0 \le 0.2\cdot3.0\text{ and }b_{eff,i}\le 0.4 = 0.380",
            ),
            ("short", r"b_{eff,i} = 0.380"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        b_i = 0.4  # M
        l_0 = 3.0  # M

        # Object to test
        form_5_7_ab_latex = Form5Dot7abFlangeEffectiveFlangeWidth(b_i=b_i, l_0=l_0).latex()

        actual = {
            "complete": form_5_7_ab_latex.complete,
            "short": form_5_7_ab_latex.short,
        }

        assert expected == actual[representation], f"{representation} representation."



"""Testing formula 5.27 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_27 import Form5Dot27EffectiveDesignModulusElasticity
from blueprints.validations import NegativeValueError


class TestForm5Dot27EffectiveDesignModulusElasticity:
    """Validation for formula 5.27 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        e_cd = 30000  # MPA
        phi_ef = 2.0  # Dimensionless

        # Object to test
        form_5_27 = Form5Dot27EffectiveDesignModulusElasticity(e_cd=e_cd, phi_ef=phi_ef)

        # Expected result, manually calculated
        manually_calculated_result = 10000.0  # MPA

        assert form_5_27 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_e_cd_is_given(self) -> None:
        """Test a negative value for e_cd."""
        # Example values
        e_cd = -30000
        phi_ef = 2.0

        with pytest.raises(NegativeValueError):
            Form5Dot27EffectiveDesignModulusElasticity(e_cd=e_cd, phi_ef=phi_ef)

    def test_raise_error_when_negative_phi_ef_is_given(self) -> None:
        """Test a negative value for phi_ef."""
        # Example values
        e_cd = 30000
        phi_ef = -2.0

        with pytest.raises(NegativeValueError):
            Form5Dot27EffectiveDesignModulusElasticity(e_cd=e_cd, phi_ef=phi_ef)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            ("complete", r"E_{cd,eff} = \frac{E_{cd}}{1 + \phi_{ef}} = \frac{30000.000}{1 + 2.000} = 10000.000"),
            ("short", "E_{cd,eff} = 10000.000"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_cd = 30000  # MPA
        phi_ef = 2.0  # Dimensionless

        # Object to test
        form_5_27_latex = Form5Dot27EffectiveDesignModulusElasticity(e_cd=e_cd, phi_ef=phi_ef).latex()

        actual = {
            "complete": form_5_27_latex.complete,
            "short": form_5_27_latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Testing formula 6.57N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_57n import Form6Dot57NNuPrime
from blueprints.validations import NegativeValueError


class TestForm6Dot57NNuPrime:
    """Validation for formula 6.57N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        f_ck = 30.0

        # Object to test
        formula = Form6Dot57NNuPrime(f_ck=f_ck)

        # Expected result, manually calculated
        manually_calculated_result = 0.88  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        "f_ck",
        [
            -30.0,  # f_ck is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, f_ck: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot57NNuPrime(f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\nu' = 1 - \frac{f_{ck}}{250} = 1 - \frac{30.000}{250} = 0.880 \ -",
            ),
            ("short", r"\nu' = 0.880 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 30.0

        # Object to test
        latex = Form6Dot57NNuPrime(f_ck=f_ck).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 3.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_1 import Form3Dot1EstimationConcreteCompressiveStrength
from blueprints.validations import NegativeValueError


class TestForm3Dot1EstimationConcreteCompressiveStrength:
    """Validation for formula 3.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = 1  # -
        f_cm = 10  # MPa
        form_3_1 = Form3Dot1EstimationConcreteCompressiveStrength(beta_cc_t=beta_cc_t, f_cm=f_cm)

        # Expected result, manually calculated
        manually_calculated_result = 10

        assert form_3_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_beta_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = -1  # -
        f_cm = 10  # MPa

        with pytest.raises(NegativeValueError):
            Form3Dot1EstimationConcreteCompressiveStrength(beta_cc_t=beta_cc_t, f_cm=f_cm)

    def test_raise_error_when_negative_f_cm_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = 1  # -
        f_cm = -10  # MPa

        with pytest.raises(NegativeValueError):
            Form3Dot1EstimationConcreteCompressiveStrength(beta_cc_t=beta_cc_t, f_cm=f_cm)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{cm}(t) = \beta_{cc}(t) \cdot f_{cm} = 1.000 \cdot 10.000 = 10.000",
            ),
            ("short", r"f_{cm}(t) = 10.000"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta_cc_t = 1  # -
        f_cm = 10  # MPa

        # Object to test
        form_3_1_latex = Form3Dot1EstimationConcreteCompressiveStrength(beta_cc_t=beta_cc_t, f_cm=f_cm).latex()

        actual = {"complete": form_3_1_latex.complete, "short": form_3_1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 4.2 of NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

import pytest

from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.formula_4_2 import Form4Dot2MinimumConcreteCover
from blueprints.validations import NegativeValueError


class TestForm4Dot1NominalConcreteCover:
    """Validation for formula 4.2 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = 0  # mm
        form_4_2 = Form4Dot2MinimumConcreteCover(
            c_min_b=c_min_b,
            c_min_dur=c_min_dur,
            delta_c_dur_gamma=delta_c_dur_gamma,
            delta_c_dur_st=delta_c_dur_st,
            delta_c_dur_add=delta_c_dur_add,
        )

        # Expected result, manually calculated
        manually_calculated_result = 15

        assert form_4_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_default_values(self) -> None:
        """Test the evaluation of the result using the default values for delta parameters."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = 20  # mm
        form_4_2 = Form4Dot2MinimumConcreteCover(
            c_min_b=c_min_b,
            c_min_dur=c_min_dur,
        )

        # Expected result, manually calculated
        manually_calculated_result = 20

        assert form_4_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_c_min_b_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = -15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = 0  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot2MinimumConcreteCover(
                c_min_b=c_min_b,
                c_min_dur=c_min_dur,
                delta_c_dur_gamma=delta_c_dur_gamma,
                delta_c_dur_st=delta_c_dur_st,
                delta_c_dur_add=delta_c_dur_add,
            )

    def test_raise_error_when_negative_c_min_dur_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = -10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = 0  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot2MinimumConcreteCover(
                c_min_b=c_min_b,
                c_min_dur=c_min_dur,
                delta_c_dur_gamma=delta_c_dur_gamma,
                delta_c_dur_st=delta_c_dur_st,
                delta_c_dur_add=delta_c_dur_add,
            )

    def test_raise_error_when_negative_delta_c_dur_gamma_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = -5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = 0  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot2MinimumConcreteCover(
                c_min_b=c_min_b,
                c_min_dur=c_min_dur,
                delta_c_dur_gamma=delta_c_dur_gamma,
                delta_c_dur_st=delta_c_dur_st,
                delta_c_dur_add=delta_c_dur_add,
            )

    def test_raise_error_when_negative_delta_c_dur_st_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = -5  # mm
        delta_c_dur_add = 0  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot2MinimumConcreteCover(
                c_min_b=c_min_b,
                c_min_dur=c_min_dur,
                delta_c_dur_gamma=delta_c_dur_gamma,
                delta_c_dur_st=delta_c_dur_st,
                delta_c_dur_add=delta_c_dur_add,
            )

    def test_raise_error_when_negative_delta_c_dur_add_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min_b = 15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = -5  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot2MinimumConcreteCover(
                c_min_b=c_min_b,
                c_min_dur=c_min_dur,
                delta_c_dur_gamma=delta_c_dur_gamma,
                delta_c_dur_st=delta_c_dur_st,
                delta_c_dur_add=delta_c_dur_add,
            )

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"c_{min} = \max \left\{c_{min,b}; c_{min,dur}+\Delta c_{dur,\gamma}-\Delta c_{dur,st}-\Delta c_{dur,add}; 10 "
                r"\text{mm}\right\} = \max \left\{15; 10+5-5-0; 10\right\} = 15.0",
            ),
            ("short", "c_{min} = 15.0"),
            (
                "string",
                r"c_{min} = \max \left\{c_{min,b}; c_{min,dur}+\Delta c_{dur,\gamma}-\Delta c_{dur,st}-\Delta c_{dur,add}; 10 "
                r"\text{mm}\right\} = \max \left\{15; 10+5-5-0; 10\right\} = 15.0",
            ),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex implementation."""
        c_min_b = 15  # mm
        c_min_dur = 10  # mm
        delta_c_dur_gamma = 5  # mm
        delta_c_dur_st = 5  # mm
        delta_c_dur_add = 0  # mm

        form = Form4Dot2MinimumConcreteCover(
            c_min_b=c_min_b,
            c_min_dur=c_min_dur,
            delta_c_dur_gamma=delta_c_dur_gamma,
            delta_c_dur_st=delta_c_dur_st,
            delta_c_dur_add=delta_c_dur_add,
        ).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Formula 6.32 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import NMM, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot32MNrdRectangular(Formula):
    r"""Class representing formula 6.32 for the calculation of [$M_{N,Rd}$]."""

    label = "6.32"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        m_pl_rd: NMM,
        n_ed: N,
        n_pl_rd: N,
    ) -> None:
        r"""[$M_{N,Rd}$] Calculation of the reduced plastic moment for rectangular solid section without fastener holes [$Nmm$].

        EN 1993-1-1:2005 art.6.2.9(3) - Formula (6.32)

        Parameters
        ----------
        m_pl_rd : NMM
            [$M_{pl,Rd}$] Plastic moment resistance of the section [$Nmm$].
        n_ed : N
            [$N_{Ed}$] Design axial force [$N$].
        n_pl_rd : N
            [$N_{pl,Rd}$] Plastic axial resistance of the section [$N$].
        """
        super().__init__()
        self.m_pl_rd = m_pl_rd
        self.n_ed = n_ed
        self.n_pl_rd = n_pl_rd

    @staticmethod
    def _evaluate(
        m_pl_rd: NMM,
        n_ed: N,
        n_pl_rd: N,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_pl_rd=n_pl_rd)
        raise_if_negative(n_ed=n_ed, m_pl_rd=m_pl_rd)

        return m_pl_rd * (1 - (n_ed / n_pl_rd) ** 2)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.32."""
        _equation: str = r"M_{pl,Rd} \cdot \left[ 1 - \left( \frac{N_{Ed}}{N_{pl,Rd}} \right)^2 \right]"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"M_{pl,Rd}": f"{self.m_pl_rd:.{n}f}",
                r"N_{Ed}": f"{self.n_ed:.{n}f}",
                r"N_{pl,Rd}": f"{self.n_pl_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                r"M_{pl,Rd}": rf"{self.m_pl_rd:.{n}f} \ Nmm",
                r"N_{Ed}": rf"{self.n_ed:.{n}f} \ N",
                r"N_{pl,Rd}": rf"{self.n_pl_rd:.{n}f} \ N",
            },
            True,
        )
        return LatexFormula(
            return_symbol=r"M_{N,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Module for testing the Formula classes."""

import pytest

from blueprints.codes.formula import ComparisonFormula, Formula


class FormulaTest(Formula):
    """Dummy formula for testing purposes."""

    label = "Dummy testing formula"
    source_document = "Dummy testing document"

    def __init__(
        self,
        first: float,
        second: float,
    ) -> None:
        """Dummy formula for testing purposes."""
        super().__init__()
        self.first = first
        self.second = second

    @staticmethod
    def _evaluate(
        first: float,
        second: float,
    ) -> float:
        """Dummy formula for testing purposes."""
        return first + second


def test_raise_error_when_changing_value_after_initialization() -> None:
    """Test that an error is raised when changing a value after initialization."""
    first = 1
    second = 2
    dummy_testing_formula = FormulaTest(first=first, second=second)
    with pytest.raises(AttributeError):
        dummy_testing_formula.first = 3


def test_raise_not_implemented_error_detailed_result() -> None:
    """Test that an error is raised when the detailed result is not implemented."""
    first = 1
    second = 2
    dummy_testing_formula = FormulaTest(first=first, second=second)
    with pytest.raises(NotImplementedError):
        _ = dummy_testing_formula.detailed_result


class ComparisonFormulaTest(ComparisonFormula):
    """Dummy comparison formula for testing purposes."""

    label = "Dummy testing comparison formula"
    source_document = "Dummy testing document"

    def __init__(
        self,
        a: float,
        b: float,
        c: float,
    ) -> None:
        """Dummy comparison formula for testing purposes."""
        super().__init__()
        self.a = a
        self.b = b
        self.c = c

    @staticmethod
    def _evaluate(
        a: float,
        b: float,
        c: float,
    ) -> float:
        """Dummy formula for testing purposes."""
        return ComparisonFormulaTest._evaluate_lhs(a=a, b=b) <= ComparisonFormulaTest._evaluate_rhs(c=c)

    @staticmethod
    def _evaluate_lhs(a: float, b: float, **_) -> float:
        """Left-hand side value of the comparison."""
        return a + b

    @staticmethod
    def _evaluate_rhs(c: float, **_) -> float:
        """Right-hand side value of the comparison."""
        return c / 2

    @property
    def unity_check(self) -> float:
        """Property to present the unity check of the formula."""
        return self.lhs / self.rhs


def test_comparison_formula_evaluation() -> None:
    """Test that the formula returns the correct result."""
    a = 10
    b = 5
    c = 40

    # check passing condition (lhs <= rhs) = True
    formula = ComparisonFormulaTest(a=a, b=b, c=c)
    assert formula

    # check failing condition (lhs > rhs) = False
    a = 30
    formula = ComparisonFormulaTest(a=a, b=b, c=c)
    assert not formula


def test_comparison_formula_lhs_property() -> None:
    """Test that the lhs property returns the correct result."""
    a = 10
    b = 5
    c = 40
    formula = ComparisonFormulaTest(a=a, b=b, c=c)

    # The lhs should be a + b = 10 + 5 = 15
    assert formula.lhs == 15


def test_comparison_formula_rhs_property() -> None:
    """Test that the rhs property returns the correct result."""
    a = 10
    b = 5
    c = 40
    formula = ComparisonFormulaTest(a=a, b=b, c=c)

    # The rhs should be c / 2 = 40 / 2 = 20
    assert formula.rhs == 20


def test_comparison_formula_unity_check_property() -> None:
    """Test that the unity check property returns the correct result."""
    a = 10
    b = 5
    c = 40
    formula = ComparisonFormulaTest(a=a, b=b, c=c)

    # The unity check should be lhs / rhs = 15 / 20 = 0.75
    assert formula.unity_check == 0.75


def test_comparison_formula_change_value_after_initialization() -> None:
    """Test that an error is raised when changing a value after initialization."""
    a = 10
    b = 5
    c = 40
    formula = ComparisonFormulaTest(a=a, b=b, c=c)

    with pytest.raises(AttributeError):
        formula.a = 30



"""Testing formula 6.39 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_39 import Form6Dot39ReducedBendingMomentResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot39ReducedBendingMomentResistance:
    """Validation for formula 6.39 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        mpl_y_rd = 5000.0
        n = 0.2
        a_w = 0.3

        formula = Form6Dot39ReducedBendingMomentResistance(mpl_y_rd=mpl_y_rd, n=n, a_w=a_w)
        manually_calculated_result = 4705.882  # Nmm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("mpl_y_rd", "n", "a_w"),
        [
            (-5000.0, 0.2, 0.3),  # mpl_y_rd is negative
            (5000.0, -0.2, 0.3),  # n is negative
            (5000.0, 0.2, -0.3),  # a_w is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, mpl_y_rd: float, n: float, a_w: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot39ReducedBendingMomentResistance(mpl_y_rd=mpl_y_rd, n=n, a_w=a_w)

    @pytest.mark.parametrize(
        ("mpl_y_rd", "n", "a_w"),
        [
            (5000.0, 0.2, 2.0),  # denominator (1 - 0.5 * a_w) <= 0
        ],
    )
    def test_raise_error_when_denominator_is_invalid(self, mpl_y_rd: float, n: float, a_w: float) -> None:
        """Test invalid denominator."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot39ReducedBendingMomentResistance(mpl_y_rd=mpl_y_rd, n=n, a_w=a_w)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{N,y,Rd} = \min \left( M_{pl,y,Rd} \cdot \frac{1 - n}{1 - 0.5 \cdot a_w}, M_{pl,y,Rd} \right) = "
                r"\min \left( 5000.000 \cdot \frac{1 - 0.200}{1 - 0.5 \cdot 0.300}, 5000.000 \right) = 4705.882 \ Nmm",
            ),
            (
                "complete_with_units",
                r"M_{N,y,Rd} = \min \left( M_{pl,y,Rd} \cdot \frac{1 - n}{1 - 0.5 \cdot a_w}, M_{pl,y,Rd} \right) = "
                r"\min \left( 5000.000 \ Nmm \cdot \frac{1 - 0.200}{1 - 0.5 \cdot 0.300}, 5000.000 \ Nmm \right) = 4705.882 \ Nmm",
            ),
            ("short", r"M_{N,y,Rd} = 4705.882 \ Nmm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        mpl_y_rd = 5000.0
        n = 0.2
        a_w = 0.3

        latex = Form6Dot39ReducedBendingMomentResistance(mpl_y_rd=mpl_y_rd, n=n, a_w=a_w).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.18 from EN 1993-1-9:2005."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_18 import Form5Dot18ComparisonGeneralSecondOrderEffects
from blueprints.validations import LessOrEqualToZeroError


class TestForm5Dot18ComparisonGeneralSecondOrderEffects:
    """Testing formula 5.18 from EN 1993-1-9:2005."""

    @pytest.mark.parametrize(
        ("f_ved", "k_1", "n_s", "length", "e_cd", "i_c", "result_manual"),
        [
            (53.3333, 0.8, 2, 5, 30_000, 100_000_000, True),
            (666.6667, 1.2, 4, 3, 35_000, 200_000_000, False),
        ],
    )
    def test_comparison(self, f_ved: float, k_1: float, n_s: float, length: float, e_cd: float, i_c: float, result_manual: bool) -> None:
        """Test the evaluation of the comparison."""
        form_5_18 = bool(Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=f_ved, k_1=k_1, n_s=n_s, length=length, e_cd=e_cd, i_c=i_c))
        assert form_5_18 == pytest.approx(expected=result_manual)

    @pytest.mark.parametrize(
        ("f_ved", "k_1", "n_s", "length", "e_cd", "i_c"),
        [
            (53.3333, 0.8, 2, 5, 30_000, 100_000_000),
            (666.6667, 1.2, 4, 3, 35_000, 200_000_000),
        ],
    )
    def test_evaluation(self, f_ved: float, k_1: float, n_s: float, length: float, e_cd: float, i_c: float) -> None:
        """Test the right-hand side evaluation of the result."""
        form_5_18 = Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=f_ved, k_1=k_1, n_s=n_s, length=length, e_cd=e_cd, i_c=i_c)
        assert form_5_18.right_hand_side == pytest.approx(expected=f_ved, abs=1e-3)

    def test_raise_error_if_negative_n_e(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for n_s."""
        n_s = -1.6  # -
        def_test = 10
        with pytest.raises(LessOrEqualToZeroError):
            bool(Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=def_test, k_1=def_test, n_s=n_s, length=def_test, e_cd=def_test, i_c=def_test))

    @pytest.mark.parametrize(
        ("f_ved", "k_1", "n_s", "length", "e_cd", "i_c", "expected_lhs_rhs"),
        [
            (53.3333, 0.8, 2, 5, 30_000, 100_000_000, 53.3333),
            (666.6667, 1.2, 4, 3, 35_000, 200_000_000, 666.6667),
        ],
    )
    def test_properties(self, f_ved: float, k_1: float, n_s: float, length: float, e_cd: float, i_c: float, expected_lhs_rhs: float) -> None:
        """Test the properties of the comparison."""
        form_5_18 = Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=f_ved, k_1=k_1, n_s=n_s, length=length, e_cd=e_cd, i_c=i_c)
        assert form_5_18.left_hand_side == pytest.approx(expected_lhs_rhs, abs=1e-3)
        assert form_5_18.right_hand_side == pytest.approx(expected_lhs_rhs, abs=1e-3)
        assert form_5_18.ratio == pytest.approx(1, abs=1e-3)

    def test_str_representation(self) -> None:
        """Test the string representation of the comparison."""
        form_5_18 = Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=53.3333, k_1=0.8, n_s=2, length=5, e_cd=30_000, i_c=100_000_000)
        expected_str = (
            r"CHECK \rightarrow F_{V,Ed} \leq \frac{n_s}{n_s + 1.6} \cdot \frac{\sum E_{cd} \cdot I_c}{L^2} \rightarrow 53.333\leq \frac{2}{3.6} "
            r"\cdot \frac{\sum 30000 \cdot 100000000}{25} \rightarrow OK"
        )
        assert str(form_5_18) == expected_str

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow F_{V,Ed} \leq \frac{n_s}{n_s + 1.6} \cdot \frac{\sum E_{cd} \cdot I_c}{L^2} "
                r"\rightarrow 53.333\leq \frac{2}{3.6} \cdot \frac{\sum 30000 \cdot 100000000}{25} \rightarrow OK",
            ),
            ("short", r"CHECK \rightarrow OK"),
        ],
    )
    def test_latex_ok(self, representation: str, expected: str) -> None:
        """Test the latex representation of the comparison."""
        # Example values
        f_ved = 53.3333
        k_1 = 0.8
        n_s = 2
        length = 5
        e_cd = 30_000
        i_c = 100_000_000

        # Object to test
        form_5_18_latex = Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=f_ved, k_1=k_1, n_s=n_s, length=length, e_cd=e_cd, i_c=i_c).latex()

        actual = {
            "complete": form_5_18_latex.complete,
            "short": form_5_18_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow F_{V,Ed} \leq \frac{n_s}{n_s + 1.6} \cdot \frac{\sum E_{cd} \cdot I_c}{L^2} \rightarrow 2000.000\leq \frac{2}{"
                r"3.6} \cdot \frac{\sum 30000 \cdot 100000000}{25} \rightarrow \text{Not OK}",
            ),
            ("short", r"CHECK \rightarrow \text{Not OK}"),
        ],
    )
    def test_latex_not_ok(self, representation: str, expected: str) -> None:
        """Test the latex representation of the comparison."""
        # Example values
        f_ved = 2000
        k_1 = 0.8
        n_s = 2
        length = 5
        e_cd = 30_000
        i_c = 100_000_000

        # Object to test
        form_5_18_latex = Form5Dot18ComparisonGeneralSecondOrderEffects(f_ved=f_ved, k_1=k_1, n_s=n_s, length=length, e_cd=e_cd, i_c=i_c).latex()

        actual = {
            "complete": form_5_18_latex.complete,
            "short": form_5_18_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 7.16a/b from EN 1992-1-1:2004: Chapter 7 - Serviceability limit state."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot16abSpanDepthRatio(Formula):
    r"""Class representing formula 7.16a/b for the calculation of the limit span/depth ratio."""

    label = "7.16a/b"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        capital_k: DIMENSIONLESS,
        f_ck: MPA,
        rho: DIMENSIONLESS,
        rho_0: DIMENSIONLESS,
        rho_prime: DIMENSIONLESS,
    ) -> None:
        r"""[$\frac{l}{d}$] Limit span/depth ratio [-].

        EN 1992-1-1:2004 art.7.4.2(2) - Formula (7.16a and 7.16b)

        Parameters
        ----------
        capital_k : DIMENSIONLESS
            [$K$] Factor to take into account the different structural systems [-].
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        rho : DIMENSIONLESS
            [$\rho$] Required tension reinforcement ratio at mid-span to resist the moment
            due to the design loads (at support for cantilevers) [-].
        rho_0 : DIMENSIONLESS
            [$\rho_0$] Reference reinforcement ratio [$\sqrt{f_{ck}} \cdot 10^{-3}$] [-].
        rho_prime : DIMENSIONLESS
            [$\rho'$] Required compression reinforcement ratio at mid-span to resist the moment
            due to design loads (at support for cantilevers) [-].
        """
        super().__init__()
        self.capital_k = capital_k
        self.f_ck = f_ck
        self.rho = rho
        self.rho_0 = rho_0
        self.rho_prime = rho_prime

    @staticmethod
    def _evaluate(
        capital_k: DIMENSIONLESS,
        f_ck: MPA,
        rho: DIMENSIONLESS,
        rho_0: DIMENSIONLESS,
        rho_prime: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(capital_k=capital_k, f_ck=f_ck, rho=rho, rho_0=rho_0, rho_prime=rho_prime)
        raise_if_less_or_equal_to_zero(rho=rho)

        if rho <= rho_0:
            l_over_d = capital_k * (11 + 1.5 * np.sqrt(f_ck) * rho_0 / rho + 3.2 * np.sqrt(f_ck) * (rho_0 / rho - 1) ** 1.5)
        else:
            l_over_d = capital_k * (11 + 1.5 * np.sqrt(f_ck) * rho_0 / (rho - rho_prime) + 1 / 12 * np.sqrt(f_ck) * np.sqrt(rho_prime / rho_0))

        return l_over_d

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.16a/b."""
        _equation: str = (
            r"\begin{cases} K \cdot \left[11 + 1.5 \cdot \sqrt{f_{ck}} \cdot \frac{\rho_0}{\rho} + "
            r"3.2 \cdot \sqrt{f_{ck}} \cdot \left(\frac{\rho_0}{\rho} - 1\right)^{3/2}\right] & \text{if } \rho \leq \rho_0 \\ "
            r"K \cdot \left[11 + 1.5 \cdot \sqrt{f_{ck}} \cdot \frac{\rho_0}{\rho - \rho'} + "
            r"\frac{1}{12} \cdot \sqrt{f_{ck}} \cdot \sqrt{\frac{\rho'}{\rho_0}}\right] & \text{if } \rho > \rho_0 \end{cases}"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"K": f"{self.capital_k:.{n}f}",
                r"f_{ck}": f"{self.f_ck:.{n}f}",
                r"\rho_0": f"{self.rho_0:.{n + 1}f}",
                r"\rho'": f"{self.rho_prime:.{n + 1}f}",
                r"\rho": f"{self.rho:.{n + 1}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\frac{l}{d}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )


class Form7Dot16ReferenceReinforcementRatio(Formula):
    r"""Class representing formula 7.16 for the calculation of [$\rho_0$]."""

    label = "7.16_rho_0"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        r"""[$\rho_0$] Reference reinforcement ratio [-].

        EN 1992-1-1:2004 art.7.4.2(2) - Formula (7.16)

        Parameters
        ----------
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck)

        return np.sqrt(f_ck) * 10**-3

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.16_rho_0."""
        _equation: str = r"\sqrt{f_{ck}} \cdot 10^{-3}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"f_{ck}": f"{self.f_ck:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\rho_0",
            result=f"{self:.6f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Formula 6.20 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot20LongitudinalShearStress(Formula):
    r"""Class representing formula 6.20 for the calculation of the longitudinal shear stress, [$v_{Ed}$]."""

    label = "6.20"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        delta_f_d: N,
        h_f: MM,
        delta_x: MM,
    ) -> None:
        r"""[$v_{Ed}$] Longitudinal shear stress [$MPa$].

        EN 1992-1-1:2004 art.6.2.4(3) - Formula (6.20)

        Parameters
        ----------
        delta_f_d : N
            [$\Delta F_{d}$] Change of the normal force in the flange over the length [$\Delta x$] [$N$].
        h_f : MM
            [$h_{f}$] Thickness of flange at the junctions [$mm$].
        delta_x : MM
            [$\Delta x$] Length under consideration, see Figure 6.7 [$mm$].
        """
        super().__init__()
        self.delta_f_d = delta_f_d
        self.h_f = h_f
        self.delta_x = delta_x

    @staticmethod
    def _evaluate(
        delta_f_d: N,
        h_f: MM,
        delta_x: MM,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(delta_f_d=delta_f_d)
        raise_if_less_or_equal_to_zero(h_f=h_f, delta_x=delta_x)

        return delta_f_d / (h_f * delta_x)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.20."""
        return LatexFormula(
            return_symbol=r"v_{Ed}",
            result=f"{self:.{n}f}",
            equation=r"\frac{\Delta F_{d}}{h_{f} \cdot \Delta x}",
            numeric_equation=rf"\frac{{{self.delta_f_d:.{n}f}}}{{{self.h_f:.{n}f} \cdot {self.delta_x:.{n}f}}}",
            comparison_operator_label="=",
            unit="MPa",
        )



"""Tests for cross-section shapes."""

import pytest
from sectionproperties.analysis import Section
from shapely import Polygon

from blueprints.structural_sections.cross_section_circle import CircularCrossSection


class TestCircularCrossSection:
    """Tests for the CircularCrossSection class."""

    def test_polygon(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the polygon property of the CircularCrossSection class."""
        assert isinstance(circular_cross_section.polygon, Polygon)

    def test_wrong_input(self) -> None:
        """Test the wrong input for the CircularCrossSection class."""
        with pytest.raises(ValueError):
            CircularCrossSection(name="Circle", diameter=-200.0, x=0.0, y=0.0)

    def test_radius(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the radius property of the CircularCrossSection class."""
        assert circular_cross_section.radius == pytest.approx(expected=100.0, rel=1e-6)

    def test_geometry_bounds(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the bounds of the geometry property."""
        bounds = circular_cross_section.polygon.bounds
        assert bounds == pytest.approx(expected=(0.0, 150.0, 200.0, 350.0), rel=1e-6)

    def test_section(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the section object of the CircularCrossSection class."""
        section = circular_cross_section.section()
        assert isinstance(section, Section)

    def test_geometry(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the geometry property of the CircularCrossSection class."""
        geometry = circular_cross_section.geometry()
        assert geometry is not None



"""Formula 6.72 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form6Dot72FatigueResistanceConcreteCompression(Formula):
    r"""Class representing formula 6.72 for checking the fatigue resistance of concrete under compression."""

    label = "6.72"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        e_cd_max_equ: MPA,
        r_equ: DIMENSIONLESS,
    ) -> None:
        r"""Check the fatigue resistance of concrete under compression.

        EN 1992-1-1:2004 art.6.8.7(1) - Formula (6.72)

        Parameters
        ----------
        e_cd_max_equ : MPA
            [$E_{cd,max,equ}$] Maximum compressive stress level [$MPa$].
        r_equ : DIMENSIONLESS
            [$R_{equ}$] Stress ratio [$-$].
        """
        super().__init__()
        self.e_cd_max_equ = e_cd_max_equ
        self.r_equ = r_equ

    @staticmethod
    def _evaluate(
        e_cd_max_equ: MPA,
        r_equ: DIMENSIONLESS,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(e_cd_max_equ=e_cd_max_equ, r_equ=r_equ)
        in_sqrt = 1 - r_equ
        raise_if_negative(in_sqrt=in_sqrt)

        return (e_cd_max_equ + 0.43 * np.sqrt(1 - r_equ)) <= 1

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.72."""
        _equation: str = r"E_{cd,max,equ} + 0.43 \cdot \sqrt{1 - R_{equ}} \leq 1"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "E_{cd,max,equ}": f"{self.e_cd_max_equ:.{n}f}",
                "R_{equ}": f"{self.r_equ:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Tests of the materials soil module."""

import pytest

from blueprints.materials.soil import SoilConsistencies, SoilConsolidationStates, SoilMainTypes, SoilMinorCompositions, SoilType


class TestSoilType:
    """Tests for the SoilType class."""

    @pytest.mark.parametrize(
        ("main_type", "minor_composition", "consolidation_state", "consistency", "description"),
        [
            (SoilMainTypes.SAND, SoilMinorCompositions.CLEAN, None, SoilConsistencies.LOOSE, None),
            (SoilMainTypes.LOAM, SoilMinorCompositions.VERY_SANDY, None, None, None),
            (SoilMainTypes.PEAT, None, SoilConsolidationStates.NORMALLY_CONSOLIDATED, None, None),
            (SoilMainTypes.PEAT, None, SoilConsolidationStates.NORMALLY_CONSOLIDATED, None, "Veen, nvt"),
        ],
    )
    def test_soil_type_init_valid_input(
        self,
        main_type: SoilMainTypes,
        minor_composition: SoilMinorCompositions,
        consolidation_state: SoilConsolidationStates | None,
        consistency: SoilConsistencies | None,
        description: str | None,
    ) -> None:
        """Tests initialization of SoilType for valid input."""
        # Giving None as input
        soil_type = SoilType(
            main_type=main_type,
            minor_composition=minor_composition,
            consolidation_state=consolidation_state,
            consistency=consistency,
            description=description,
        )

        assert soil_type

    @pytest.mark.parametrize(
        ("main_type", "minor_composition", "consolidation_state", "consistency", "expected_default_description"),
        [
            (SoilMainTypes.CLAY, None, None, None, "Clay"),
            (SoilMainTypes.LOAM, SoilMinorCompositions.VERY_SANDY, None, None, "Loam, Very sandy"),
            (SoilMainTypes.SAND, SoilMinorCompositions.CLEAN, None, SoilConsistencies.LOOSE, "Sand, Clean, Loose"),
            (SoilMainTypes.PEAT, None, SoilConsolidationStates.NORMALLY_CONSOLIDATED, None, "Peat, Normally consolidated"),
        ],
    )
    def test_soil_type_default_description(
        self,
        main_type: SoilMainTypes,
        minor_composition: SoilMinorCompositions,
        consolidation_state: SoilConsolidationStates | None,
        consistency: SoilConsistencies | None,
        expected_default_description: str,
    ) -> None:
        """Tests the property method `default_description` of SoilType."""
        soil_type = SoilType(
            main_type=main_type,
            minor_composition=minor_composition,
            consolidation_state=consolidation_state,
            consistency=consistency,
        )

        assert soil_type.default_description == expected_default_description



"""Testing sub-formula 8.16 (α2) of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_16 import (
    SubForm8Dot16Alpha2,
)


class TestSubForm8Dot16Alpha2:
    """Validation for sub-formula 8.16 (α2) from EN 1992-1-1:2004."""

    def test_evaluation_circular(self) -> None:
        """Test the evaluation of the result when type_of_wire is circular."""
        # example values
        type_of_wire = "circular"

        sub_form_8_16_alpha_2 = SubForm8Dot16Alpha2(type_of_wire=type_of_wire)

        # manually calculated result
        manually_calculated_result = 0.25  # [-]

        assert sub_form_8_16_alpha_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_3_7_wire_strand(self) -> None:
        """Test the evaluation of the result when type_of_wire is 3_7_wire_strands."""
        # example values
        type_of_wire = "3_7_wire_strands"

        sub_form_8_16_alpha_2 = SubForm8Dot16Alpha2(type_of_wire=type_of_wire)

        # manually calculated result
        manually_calculated_result = 0.19

        assert sub_form_8_16_alpha_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_invalid_type_of_wire(self) -> None:
        """Test that a ValueError is raised when an invalid value is passed for type_of_wire."""
        type_of_wire = "invalid"

        with pytest.raises(ValueError):
            SubForm8Dot16Alpha2(type_of_wire=type_of_wire)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (r"\alpha_2 \rightarrow type\;of\;wire \rightarrow circular \rightarrow 0.25"),
            ),
            ("short", r"\alpha_2 \rightarrow 0.25"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        type_of_wire = "circular"

        # Object to test
        form_8_16_p2_latex = SubForm8Dot16Alpha2(type_of_wire=type_of_wire).latex()

        actual = {"complete": form_8_16_p2_latex.complete, "short": form_8_16_p2_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 7.7 from EN 1995-1-1:2004."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1995_1_1_2004.chapter_7_serviceability_limit_states.formula_7_7 import Form7Dot7NumberOfFOVibrations
from blueprints.validations import LessOrEqualToZeroError


class TestForm7Dot7NumberOfFOVibrations:
    """Validation for formula 7.7 from EN 1995-1-1:2004."""

    @pytest.mark.parametrize(
        ("f_1", "b", "length", "ei_l", "ei_b", "expected_result"),
        [
            (8.96, 1.0, 5.3, 5.81e12, 3.89e11, 0.774),
            (28.2843, 1.0, 1.0, 1.0001, 1, 1.00),
        ],
        ids=["decimal result", "result to 1"],
    )
    def test_evaluation(self, f_1: float, b: float, length: float, ei_l: float, ei_b: float, expected_result: float) -> None:
        """Test the evaluation of the result."""
        form_7_7 = Form7Dot7NumberOfFOVibrations(f_1=f_1, b=b, length=length, ei_l=ei_l, ei_b=ei_b)
        assert form_7_7 == pytest.approx(expected=expected_result, rel=1e-3)

    @pytest.mark.parametrize(
        ("f_1", "b", "length", "ei_l", "ei_b", "expectation"),
        [
            (8.96, 1.0, 5.3, 5.81e12, 3.89e11, does_not_raise()),
            (20, 1.0, 1.0, 5.82e12, 5.81e12, does_not_raise()),
            (41, 1, 1, 2, 1, pytest.raises(ValueError, match="exceeds the allowed limit of 40.")),
            (20, 1, 1, 1, 1, pytest.raises(ValueError, match="must be bigger than")),
            (20, 1, 1, 1, 2, pytest.raises(ValueError, match="must be bigger than")),
            (20, 1, 1, 2, 0, pytest.raises(LessOrEqualToZeroError)),
            (20, 0, 1, 2, 2, pytest.raises(LessOrEqualToZeroError)),
            (20, 1, -20, 2, 3, pytest.raises(LessOrEqualToZeroError)),
            (20, 1, 1, -2, 3, pytest.raises(LessOrEqualToZeroError)),
        ],
        ids=["passes decimal result", "passes integer", "f_1 > 40", "ei_b = ei_l", "ei_b > ei_l", "ei_b = 0", "b = 0", "length < 0", "ei_l < 0"],
    )
    def test_raise_error_incorrect_args(
        self, f_1: float, b: float, length: float, ei_l: float, ei_b: float, expectation: AbstractContextManager
    ) -> None:
        """Test if errors are raised."""
        with expectation:
            assert Form7Dot7NumberOfFOVibrations(f_1=f_1, b=b, length=length, ei_l=ei_l, ei_b=ei_b) is not None

    @pytest.mark.parametrize(
        ("f_1", "b", "length", "ei_l", "ei_b", "rep_short", "rep_long"),
        [
            (
                8.96,  # HZ
                1.0,  # M
                5.3,  # M
                5.81e7,  # NM2_M
                3.89e6,  # NM2_M
                r"n_{40} = 0.77",
                r"n_{40} = \left\{\left ( \left (\frac{40}{f_1} \right )^2 -1 \right)\left ( \frac{b}{l } \right )^4 \frac{(EI)_l}{(EI)_b}\right\}^{0.25} = \left\{\left ( \left (\frac{40}{8.96} \right )^2 -1 \right)\left ( \frac{1.00}{5.30} \right )^4 \frac{58100000.00}{3890000.00}\right\}^{0.25} = 0.77",  # noqa: E501
            ),
            (
                28.2843,  # HZ
                1,  # M
                1,  # M
                1.001,  # NM2_M
                1,  # NM2_M
                r"n_{40} = 1.00",
                r"n_{40} = \left\{\left ( \left (\frac{40}{f_1} \right )^2 -1 \right)\left ( \frac{b}{l } \right )^4 \frac{(EI)_l}{(EI)_b}\right\}^{0.25} = \left\{\left ( \left (\frac{40}{28.28} \right )^2 -1 \right)\left ( \frac{1.00}{1.00} \right )^4 \frac{1.00}{1.00}\right\}^{0.25} = 1.00",  # noqa: E501
            ),
        ],
        ids=["test with decimal output", "test with integer output"],
    )
    def test_latex(self, f_1: float, b: float, length: float, ei_l: float, ei_b: float, rep_short: str, rep_long: str) -> None:
        """Test the latex representation of the formula."""
        form_7_7_latex = Form7Dot7NumberOfFOVibrations(f_1=f_1, b=b, length=length, ei_l=ei_l, ei_b=ei_b).latex()
        assert form_7_7_latex.complete == rep_long
        assert form_7_7_latex.short == rep_short



"""Module to keep track of type aliases used in Blueprints."""

# <editor-fold desc="COUNTERS">
AMOUNT = int
"""Amount, represented as a int."""
# </editor-fold>

# <editor-fold desc="DISTANCES">
MM = float
"""Millimeters (mm), represented as a float."""
CM = float
"""Centimeters (cm), represented as a float."""
M = float
"""Meters (m), represented as a float."""
KM = float
"""Kilometers (km), represented as a float."""
# </editor-fold>

# <editor-fold desc="ROTATIONS">
DEG = float
"""Degrees (deg), represented as a float."""
RAD = float
"""Radians (rad), represented as a float."""
# </editor-fold>

# <editor-fold desc="WEIGHTS">
G = float
"""Grams (g), represented as a float."""
KG = float
"""Kilograms (kg), represented as a float."""
KG_M3 = float
"""Kilograms per cubic meter (kg/m³), represented as a float."""
KG_M2 = float
"""Kilograms per square meter (kg/m2), represented as a float."""
KG_M = float
"""Kilograms per meter (kg/m), represented as a float."""
# </editor-fold>

# <editor-fold desc="FORCES">
N = float
"""Newtons (N), represented as a float."""
KN = float
"""Kilonewtons (kN), value represented as a float."""
# </editor-fold>

# <editor-fold desc="FORCE PER UNIT LENGTH">
N_MM = float
"""Newtons per millimeter (N/mm), represented as a float."""
N_M = float
"""Newtons per meter (N/m), represented as a float."""
KN_M = float
"""Kilonewtons per meter (kN/m), represented as a float."""
# </editor-fold>

# <editor-fold desc="MOMENTS">
NMM = float
"""Newton-millimeters (Nmm), represented as a float."""
NM = float
"""Newton-meters (Nm), represented as a float."""
KNM = float
"""Kilonewton-meters (kNm), represented as a float."""
# </editor-fold>

# <editor-fold desc="STRESSES">
KPA = float
"""Kilopascals (KPa), represented as a float."""
MPA = float
"""Megapascals (MPa), represented as a float."""
GPA = float
"""Gigapascals (GPa), represented as a float."""
# </editor-fold>

# <editor-fold desc="AREAS">
MM2 = float
"""Square millimeters (mm²), represented as a float."""
MM2_M = float
"""Square millimeters per meter, (mm²/m), represented as a float."""
CM2 = float
"""Square centimeters (cm²), represented as a float."""
DM2 = float
"""Square decimeters (dm²), represented as a float."""
M2 = float
"""Square meters (m²), represented as a float."""
KM2 = float
"""Square kilometers (km²), represented as a float."""
# </editor-fold>

# <editor-fold desc="VOLUMES">
MM3 = float
"""Cubic millimeters (mm³), represented as a float."""
CM3 = float
"""Cubic centimeters (cm³), represented as a float."""
M3 = float
"""Cubic meters (m³), represented as a float."""
M3_M = float
"""Cubic meters per meter (m³/m), represented as a float."""
# </editor-fold>

# <editor-fold desc="BI-QUADRATIC">
MM4 = float
"""Millimeters to the fourth power (mm⁴), represented as a float."""
CM4 = float
"""Centimeters to the fourth power (cm⁴), represented as a float."""
M4 = float
"""Meters to the fourth power (m⁴), represented as a float."""
# </editor-fold>

# <editor-fold desc="COMPOUNDS">
N_M2 = float
"""Newton per square meter (N/m²), represented as a float."""
KN_M2 = float
"""Kilonewton per square meter (kN/m²), represented as a float."""
KN_M3 = float
"""Kilonewton per cubic meter (kN/m³), represented as a float."""
# </editor-fold>

# <editor-fold desc="TIME">
SECONDS = float
"""Seconds (sec), represented as a float."""
MINUTES = float
"""Minutes (min), represented as a float."""
HOURS = float
"""Hours (h), represented as a float."""
DAYS = float
"""Days (days), represented as a float."""
WEEKS = float
"""Weeks (weeks), represented as a float."""
MONTHS = float
"""Months (months), represented as a float."""
YEARS = float
"""Years (years), represented as a float."""
# </editor-fold>

# <editor-fold desc="Percentages">
PERCENTAGE = float
"""Percentage (%), represented as a float."""
PER_MILLE = float
"""Per mille (‰), represented as a float."""
# </editor-fold>

# <editor-fold desc="RATIOS">
RATIO = float
"""Ratio, represented as a float."""
PER_DEGREE = float
"""Per degree (1/°), represented as a float."""
MM_KN = float
"""Ratio of deflection over applied load."""
DIMENSIONLESS = float
"""Dimensionless, represented as a float."""
# </editor-fold>

# <editor-fold desc="FREQUENCY">
HZ = float
"""Frequency (hertz), represented as a float."""
# </editor-fold>

# <editor-fold desc="TEMPERATURE">
DEG_C = float
"""Degrees (Celsius), represented as a float."""
DEG_K = float
"""Degrees (Kelvin), represented as a float."""
# </editor-fold>

# <editor-fold desc="RECIPROCAL">
ONE_OVER_MM = float
"""Reciprocal (1/mm), represented as a float."""
ONE_OVER_M = float
"""Reciprocal (1/m), represented as a float."""
# </editor-fold>

# <editor-fold desc="VELOCITIES">
M_PER_S = float
"""Velocity (m/s), represented as a float."""
KM_PER_H = float
"""Velocity (km/h), represented as a float."""
M_NS2 = float
"""Velocity response of unit impulse load (m/Ns²), represented as a float."""
# </editor-fold>

# <editor-fold desc="ACCELERATIONS">
M_PER_S2 = float
"""Acceleration (m/s²), represented as a float."""
MM_PER_S2 = float
"""Acceleration (mm/s²), represented as a float."""
# </editor-fold>

# <editor-fold desc="STIFFNESSES">
NM2_M = float
"""Bending stiffness (Nm²/m), represented as a float."""
# </editor-fold>



"""Generate downloadable Python code files from Jupyter notebooks."""

import json
from pathlib import Path

import mkdocs_gen_files


def extract_code_from_notebook(notebook_path: Path) -> str:
    """Extract all Python code cells from a Jupyter notebook.

    Parameters
    ----------
    notebook_path : Path
        Path to the notebook file.

    Returns
    -------
    str
        Combined Python code from all code cells.
    """
    with open(notebook_path, encoding="utf-8") as f:
        notebook = json.load(f)

    code_lines = []

    # Add a header comment
    code_lines.append(f"# Code automatically extracted from {notebook_path.name}.\n")

    for cell in notebook.get("cells", []):
        if cell.get("cell_type") == "code":
            source = cell.get("source", [])
            if source:
                # Join source lines if it's a list, otherwise use as is
                cell_code = "".join(source) if isinstance(source, list) else source

                # Skip empty cells or cells with only whitespace
                if cell_code.strip():
                    code_lines.append(cell_code)
                    # Add spacing between cells
                    if not cell_code.endswith("\n"):
                        code_lines.append("\n")
                    code_lines.append("\n")

    # removes trailing whitespace and ensures a single newline at the end
    return "".join(code_lines).rstrip() + "\n"


def main() -> None:
    """Generate Python code files from all Jupyter notebooks in docs."""
    root = Path(__file__).parents[2]
    docs_dir = root / "docs"

    # Find all notebook files recursively in docs
    notebook_files = list(docs_dir.rglob("*.ipynb"))

    for notebook_path in notebook_files:
        # Extract code from notebook
        code_content = extract_code_from_notebook(notebook_path)

        # Calculate relative path from docs directory
        relative_path = notebook_path.relative_to(docs_dir)

        # Generate corresponding .py file path preserving directory structure
        py_filename = notebook_path.stem + ".py"
        output_path = relative_path.parent / notebook_path.stem / py_filename

        # Write the extracted code to a .py file
        with mkdocs_gen_files.open(output_path, "w") as f:
            f.write(code_content)

        print(f"Generated {output_path} from {relative_path}")


main()



"""Formula 5.8 from EN 1993-5:2007 Chapter 5 - Ultimate limit states."""

import numpy as np

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction
from blueprints.type_alias import DIMENSIONLESS, MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot8RelativeWebSlenderness(Formula):
    """Class representing formula 5.8 for relative slenderness of the web."""

    label = "5.8"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        c: MM,  # Length of the web
        t_w: MM,  # Thickness of the web
        f_y: MPA,  # Yield strength
        e: MPA,  # Young's modulus
    ) -> None:
        r"""[$\overline{\lambda}$] Calculate the relative slenderness of the web [$-$].

        EN 1993-5:2007(E) art.5.2.2(7) - Formula (5.8)

        Parameters
        ----------
        c : MM
            [$c$] Length of the web in [$mm$].
        t_w : MM
            [$t_{w}$] Thickness of the web in [$mm$].
        f_y : MPA
            [$f_{y}$] Yield strength in [$MPa$].
        e : MPA
            [$E$] Young's modulus in [$MPa$].
        """
        super().__init__()
        self.c: MM = c
        self.t_w: MM = t_w
        self.f_y: MPA = f_y
        self.e: MPA = e

    @staticmethod
    def _evaluate(
        c: MM,
        t_w: MM,
        f_y: MPA,
        e: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula for relative slenderness."""
        raise_if_less_or_equal_to_zero(c=c, t_w=t_w, f_y=f_y, e=e)
        return 0.346 * (c / t_w) * np.sqrt(f_y / e)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 5.8."""
        return LatexFormula(
            return_symbol=r"\overline{\lambda}",
            result=f"{self:.{n}f}",
            equation=rf"0.346 \cdot {latex_fraction(numerator='c', denominator='t_w')} \sqrt{{{latex_fraction(numerator='f_y', denominator='E')}}}",
            numeric_equation=rf"0.346 \cdot {latex_fraction(numerator=self.c, denominator=self.t_w)} \sqrt{{"
            rf"{latex_fraction(numerator=self.f_y, denominator=self.e)}}}",
            comparison_operator_label="=",
        )



"""Testing formula 3.3 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_3 import (
    Form3Dot3AxialTensileStrengthFromTensileSplittingStrength,
)


class TestForm3Dot3AxialTensileStrengthFromTensileSplittingStrength:
    """Validation for formula 3.3 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ct_sp = 3.4  # MPa
        form_3_3 = Form3Dot3AxialTensileStrengthFromTensileSplittingStrength(f_ct_sp=f_ct_sp)

        # Expected result, manually calculated
        manually_calculated_result = 3.06

        assert form_3_3 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_ct_sp_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ct_sp = -3.4  # MPa

        with pytest.raises(ValueError):
            Form3Dot3AxialTensileStrengthFromTensileSplittingStrength(f_ct_sp=f_ct_sp)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ct} = 0.9 \cdot f_{ct,sp} = 0.9 \cdot 3.400 = 3.060",
            ),
            ("short", r"f_{ct} = 3.060"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ct_sp = 3.4  # MPa

        # Object to test
        form_3_3_latex = Form3Dot3AxialTensileStrengthFromTensileSplittingStrength(f_ct_sp=f_ct_sp).latex()

        actual = {"complete": form_3_3_latex.complete, "short": form_3_3_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 9.11 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_11 import Form9Dot11MinimumShearReinforcement
from blueprints.validations import GreaterThan90Error, LessOrEqualToZeroError, NegativeValueError


class TestForm9Dot11MinimumShearReinforcement:
    """Validation for formula 9.11 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        alpha = 45.0
        s_r = 200.0
        s_t = 200.0
        f_ck = 30.0
        f_yk = 500.0

        # Object to test
        formula = Form9Dot11MinimumShearReinforcement(alpha=alpha, s_r=s_r, s_t=s_t, f_ck=f_ck, f_yk=f_yk)

        # Expected result, manually calculated
        manually_calculated_result = 19.8296747326  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("alpha", "s_r", "s_t", "f_ck", "f_yk"),
        [
            (-45.0, 200.0, 200.0, 30.0, 500.0),  # alpha is negative
            (45.0, -200.0, 200.0, 30.0, 500.0),  # s_r is negative
            (45.0, 200.0, -200.0, 30.0, 500.0),  # s_t is negative
            (45.0, 200.0, 200.0, -30.0, 500.0),  # f_ck is negative
            (45.0, 200.0, 200.0, 30.0, -500.0),  # f_yk is negative
            (45.0, 0.0, 200.0, 30.0, 500.0),  # s_r is zero
            (45.0, 200.0, 0.0, 30.0, 500.0),  # s_t is zero
            (45.0, 200.0, 200.0, 30.0, 0.0),  # f_yk is zero
            (91.0, 200.0, 200.0, 30.0, 500.0),  # alpha is greater than 90
        ],
    )
    def test_raise_error_when_invalid_values_are_given_zero_or_negative(self, alpha: float, s_r: float, s_t: float, f_ck: float, f_yk: float) -> None:
        """Test invalid values that are zero or negative."""
        with pytest.raises((LessOrEqualToZeroError, NegativeValueError, GreaterThan90Error)):
            Form9Dot11MinimumShearReinforcement(alpha=alpha, s_r=s_r, s_t=s_t, f_ck=f_ck, f_yk=f_yk)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_{sw,min} = \frac{0.08 \cdot \sqrt{f_{ck}}}{f_{yk}} \cdot \frac{s_r \cdot s_t}"
                r"{1.5 \cdot \sin(\alpha) + \cos(\alpha)} = "
                r"\frac{0.08 \cdot \sqrt{30.000}}{500.000} \cdot \frac{200.000 \cdot 200.000}{1.5 \cdot \sin(45.000) "
                r"+ \cos(45.000)} = 19.830 \ mm^2",
            ),
            ("short", r"A_{sw,min} = 19.830 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        alpha = 45.0
        s_r = 200.0
        s_t = 200.0
        f_ck = 30.0
        f_yk = 500.0

        # Object to test
        latex = Form9Dot11MinimumShearReinforcement(alpha=alpha, s_r=s_r, s_t=s_t, f_ck=f_ck, f_yk=f_yk).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."
